
# show the charts and tables
import datetime as dt
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
from typing import List, Tuple, Union
import os

from pandas.core.arrays import boolean

def display_accuracy_chart(y_actual:Union[np.array, List[float], pd.Series], 
                           y_pred:Union[np.array, List[float], pd.Series], 
                           y_label:str=None, 
                           accuracy:Tuple[str, str]=None) -> None:
    # set up figure and subplots
    fig, ax = plt.subplots(figsize=(14,8), nrows=2, ncols=1, gridspec_kw={'height_ratios': [3, 1]})
    
    # create plot df
    plot_df = pd.DataFrame()
    plot_df['Actual'] = y_actual
    plot_df['Predicted'] = y_pred
    plot_df['Error'] = (y_pred - y_actual) / y_actual * 100
    
    # plot actual vs predicted on grid
    if y_label:
        plot_df[['Actual', 'Predicted']].plot(ax=ax[0], ylabel=y_label)
    else:
        plot_df[['Actual', 'Predicted']].plot(ax=ax[0])

    if accuracy:
        ax[0].annotate(f'{accuracy[0]} = {accuracy[1]}', xy=(0.05, 0.92), xycoords='axes fraction')

    ax[0].legend(loc="upper center", bbox_to_anchor=(0.5, 1.12), ncol=2)
    ax[0].grid(True, which='both')
    
    # plot error on grid
    plot_df[['Error']].plot(ax=ax[1], color='red')
    ax[1].grid(True, which='both')
    ax[1].legend(loc="upper center", bbox_to_anchor=(0.5, 1.35), ncol=2)
    fmt = '%.0f%%' # Format you want the ticks, e.g. '40%'
    yticks = mtick.FormatStrFormatter(fmt)
    ax[1].yaxis.set_major_formatter(yticks)
    
    # show plots
    fig.autofmt_xdate(rotation=45)
    plt.gcf().suptitle("Actual vs Predicted", fontsize=20)
    
    plt.show()

def display_contrib_chart(pred_df:pd.DataFrame) -> None:
    f = plt.figure(figsize=(20,10))
    plt.title('Contribution Chart', color='black', fontsize=20, y=1.03)
    (pred_df.sum() / pred_df.sum().sum()).plot(kind='barh')
    plt.show()

def display_decomp_chart(pred_df:pd.DataFrame) -> None:
    plot_df = pred_df.copy()
    
    # separate positive & negative values
    plot_df = plot_df[plot_df>=0].join(plot_df[plot_df<0], lsuffix="", rsuffix="_$negative")
    plot_df.rename(columns=lambda x: '-'+x.replace('_$negative','') if '_$negative' in x else x, inplace=True)
    plot_df.dropna(axis=1, how='all', inplace=True)
    
    # plot area chart
    f = plt.figure()
    plt.title('Decomposition Chart', color='black', fontsize=20, y=1.03)
    plot_df.plot(kind='area', figsize=(20,10), ax=f.gca())
    plt.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))

    plt.show()

def save_model(y_label:str, 
               error:Union[np.array, List[float], pd.Series], 
               X_labels:List[str], 
               coefficients:Union[np.array, List[float], pd.Series], 
               algo:str="LinearRegression", file_loc:str=None) -> None:
    
    if file_loc is None:
        file_loc = "../results/models.csv"

    timestamp = dt.datetime.today().strftime('%Y-%m-%d %H:%M')

    data = {
        'timestamp': timestamp,
        'y_label': y_label,
        'error': error,
        'X_labels': X_labels,
        'coefficients': [round(c, 3) for c in coefficients],
        'algo': algo,
    }
    df = pd.DataFrame([data])

    if os.path.isfile(file_loc):
        models = pd.read_csv(file_loc)
        df = models.append(df)
    
    model_no = df.shape[0]
    df.to_csv(file_loc, index=False)
    print(f"Model {model_no} saved to {file_loc}")

def load_models(file_loc:str=None) -> pd.DataFrame:
    if file_loc is None:
        file_loc = "../results/models.csv"

    if os.path.isfile(file_loc):
        models = pd.read_csv(file_loc)
        return models
    else:
        print(f"No models saved at {file_loc}")
        return pd.DataFrame(columns=['timestamp', 'y_label', 'X_labels', 'algo', 'error'])

def clear_models(file_loc:str=None) -> None:
    if file_loc is None:
        file_loc = "../results/models.csv"

    if os.path.isfile(file_loc):
        os.remove(file_loc)

        print(f"Models cleared from {file_loc}")

    else:
        print(f"No models saved at {file_loc}")


import numpy as np
import pandas as pd
from sklearn import metrics
import statsmodels.stats.api as sms
from statsmodels.compat import lzip
from typing import List, Tuple, Union

def calculate_r2(y_actual: Union[np.array, List[float], pd.Series], 
                 y_pred: Union[np.array, List[float], pd.Series]) -> float:
    # r squared score
    return round(metrics.r2_score(y_actual, y_pred),3)

def calculate_mae(y_actual: Union[np.array, List[float], pd.Series], 
                  y_pred: Union[np.array, List[float], pd.Series]) -> float:
    # mean absolute error
    return round(metrics.mean_absolute_error(y_actual, y_pred),3)

def calculate_mape(y_actual: Union[np.array, List[float], pd.Series], 
                   y_pred: Union[np.array, List[float], pd.Series]) -> float:
    # mean absolute percentage error
    return round(metrics.mean_absolute_error(y_actual, y_pred)/np.sum(y_actual),3)

def calculate_mse(y_actual: Union[np.array, List[float], pd.Series], 
                  y_pred: Union[np.array, List[float], pd.Series]) -> float:
    # mean squared error
    return round(metrics.mean_squared_error(y_actual, y_pred),3)

def calculate_rmse(y_actual: Union[np.array, List[float], pd.Series], 
                   y_pred: Union[np.array, List[float], pd.Series]) -> float:
    # root mean squared error
    return round(np.sqrt(metrics.mean_squared_error(y_actual, y_pred)),3)

def calculate_nrmse(y_actual: Union[np.array, List[float], pd.Series], 
                    y_pred: Union[np.array, List[float], pd.Series]) -> float:
    # normalized root mean square error
    return round(np.sqrt(metrics.mean_squared_error(y_actual, y_pred))/y_actual.mean(),3)

def run_jarque_bera_tests(residuals: Union[np.array, List[float], pd.Series]) -> List[Tuple[str, float]]:
    # Tests for normality of the residuals
    # skewness should be between -2 and 2
    # kurtosis should be between -7 and 7
    name = ['Jarque-Bera', 'Chi^2 prob', 'Skewness', 'Kurtosis']
    test = sms.jarque_bera(residuals)

    return lzip(name, test)

def run_breuschpagan_tests(residuals: Union[np.array, List[float], pd.Series], 
                           exog) -> List[Tuple[str, float]]:
    # tests for heteroskedasticity
    # p-value should be less than 0.05
    name = ['Lagrange', 'p-value','f-value', 'f p-value']
    test = sms.het_breuschpagan(residuals, exog)

    return lzip(name, test)

def run_goldfeldquandt_tests(residuals: Union[np.array, List[float], pd.Series], 
                             exog) -> List[Tuple[str, float]]:
    # tests for heteroskedasticity
    # p-value should be less than 0.05
    name = ['F statistic', 'p-value']
    test = sms.het_breuschpagan(residuals, exog)

    return lzip(name, test)

# # need a way to run without passing results object
# def run_harvey_collier_tests(residuals: Union[np.array, List[float], pd.Series], exog):
#     # p-value should be less than 0.05
#     name = ['t value', 'p value']
#     test = sms.linear_harvey_collier(results)

#     return lzip(name, test)

def run_condition_number_test(exog):
    # tests for multicollinearity
    # condition no should be less than 30
    condition = np.linalg.cond(exog)

def run_ljungbox_tests(residuals: Union[np.array, List[float], pd.Series], 
                       X_labels:List[str]) -> List[Tuple[str, float]]:
    # tests for autocorrelation
    # p-value should be less than 0.05
    name = ['Ljung-Box stat', 'p-value']
    lags = min(len(X_labels)/2-2, 40)
    test = sms.acorr_ljungbox(residuals, lags=[lags])
    return lzip(name, test)

def run_durbin_watson_test(residuals):
    # tests for autocorrelation
    # durbin watson should be between 1.5 and 2.5
    test = sms.durbin_watson(residuals)
    return ('Durbin Watson', test)

# # need a way to run without passing results object
# def run_rainbox_tests(residuals: Union[np.array, List[float], pd.Series], 
#                       X_labels: List[str]):
#     # tests for linearity
#     # p-value should be less than 0.05
#     name = ['rainbow F stat', 'rainbow F stat p-value']
#     test = sms.linear_rainbow(results)
#     return lzip(name, test)


# feature selection

import numpy as np
import pandas as pd
from statsmodels.stats.outliers_influence import variance_inflation_factor as vif
from sklearn.feature_selection import f_regression
np.seterr(divide='ignore', invalid='ignore') # hide error warning for vif
from sklearn.feature_selection import f_regression, RFE
from sklearn.linear_model import LinearRegression
from typing import List, Tuple, Union

from .build import run_regression
from .validate import calculate_nrmse

def guess_date_column(df:pd.DataFrame) -> None:
    guesses = ['date', 'Date', 'day', 'Day', 'week', 'Week', 'Month', 'month']
    columns = df.columns
    for x in guesses:
        if x in columns:
            return x
    return None

def guess_y_column(df):
    guesses = ['revenue', 'sales', 'conversions', 'purchases']
    columns = [x.lower() for x in df.columns]
    for x in guesses:
        if x in columns:
            return x
    return None

def guess_media_columns(df):
    guesses = ['cost', 'spend', 'impression', 'spent', 'clicks']
    columns = [x.lower() for x in df.columns]
    media_columns = []
    for x in guesses:
        for y in columns:
            if x in y:
                media_columns.append(y)

    return media_columns

def add_X_labels(X_labels, add_cols):
    for x in add_cols:
        if x not in X_labels:
            X_labels.append(x)

    return X_labels

def del_X_labels(X_labels:List[str], del_cols:List[str]) -> List[str]:
    for x in del_cols:
        if x in X_labels:
            X_labels.remove(x)

    return X_labels

def get_all_X_labels(df:pd.DataFrame, y_label:str, date_label:str=None) -> List[str]:
    X_labels = list(df.columns)
    X_labels.remove(y_label)
    if date_label:
        X_labels.remove(date_label)
    
    return X_labels

def get_cols_containing(df:pd.DataFrame, containing:str) -> List[str]:
    return [x for x in list(df.columns) if containing in x]

def y_variable_correlation(df:pd.DataFrame, y_label:str, X_labels:List[str], min_corr:float=0.3) -> Tuple[List, pd.DataFrame]:
    # 1.0 = Perfect
    # 0.7 = Strong
    # 0.5 = Moderate
    # 0.3 = Weak
    # 0 = None
    
    all_variables = X_labels.copy()
    all_variables.extend([y_label])

    corr = df[all_variables].corr()
    corr_df = pd.DataFrame({'corr':abs(corr[y_label].drop(y_label))})
    corr_df['corr_keep'] = corr_df['corr'] > min_corr
    
    corr_keep = list(corr_df[corr_df['corr_keep']==True].index.values)
    
    return corr_keep, corr_df

def variance_inflation_factor(df:pd.DataFrame, X_labels:List[str], max_vif:int=5) -> Tuple[List, pd.DataFrame]:
    # Variance Inflation Factor (VIF)
    # tests for colinearity: A VIF of over 10 for some feature indicates that over 90% 
    # of the variance in that feature is explained by the remaining features. Over 100 
    # indicates over 99%. Best practice is to keep variables with a VIF less than 5.

    X = df[X_labels]
    X_np = np.array(X)

    vif_results = [(X.columns[i], vif(X_np, i)) for i in range(X_np.shape[1])]
    vif_df = pd.DataFrame(vif_results)
    vif_df.columns = ['idx', 'vif']
    vif_df.index = vif_df['idx']
    vif_df.drop(['idx'], axis=1, inplace=True)
    vif_df.index.name = None
    vif_df['vif_keep'] = vif_df['vif'] < max_vif
    
    vif_keep = list(vif_df[vif_df['vif_keep']==True].index.values)

    return vif_keep, vif_df
    
def backwards_feature_elimination(df:pd.DataFrame, y_label:str, 
                                  X_labels:List[str], max_p:float=0.05) -> Tuple[List, pd.DataFrame]:
    # backwards feature elimination
    X_labels_left = X_labels.copy()
    while (len(X_labels_left)>0):
        p_values = run_regression(df, y_label, X_labels_left)[3]
        p = pd.Series(p_values,index=X_labels_left)      
        pmax = max(p)
        feature_with_p_max = p.idxmax()
        if (pmax > max_p):
            X_labels_left.remove(feature_with_p_max)
        else:
            break
    
    bfe = pd.Series(X_labels).apply(lambda x: x in X_labels_left)
    bfe.index = X_labels
    bfe_df = pd.DataFrame({'bfe_keep': bfe})
    bfe_keep = list(bfe_df[bfe_df['bfe_keep']==True].index.values)

    return bfe_keep, bfe_df

def recursive_feature_elimination(df:pd.DataFrame, y_label:str, X_labels:List[str], max_features:List[float]=None) -> Tuple[pd.Series, pd.DataFrame]:
    if max_features is None:
        max_features = max(round(len(X_labels)/5),1)

    rfe = RFE(LinearRegression(), n_features_to_select=max_features).fit(df[X_labels], df[y_label])
    rfe_keep = pd.Series(rfe.support_)
    rfe_keep.index = X_labels
    
    rfe_df = pd.DataFrame({'rfe_keep': rfe_keep})
    rfe_df['rfe_ranking'] = rfe.ranking_
    return rfe_keep, rfe_df

def find_best_feature(df, y_label, X_candidates, X_labels=None):
    models = dict()
    if X_labels is None:
        X_labels = []

    for x in X_candidates:
        y_actual, y_pred, _ = run_regression(df, y_label, [x] + X_labels)

        # test accuracy
        nrmse = calculate_nrmse(y_actual, y_pred)
        
        models[x] = nrmse

    models_df = pd.DataFrame(models.items(), columns=['candidate', 'nrmse'])
    models_df.index = models_df['candidate']
    models_df.drop(columns=['candidate'], inplace=True)
    min_label = models_df[['nrmse']].idxmin()[0]
    return min_label

# TODO: eliminate negative coefficients



import numpy as np
import pandas as pd
import statsmodels.api as sm
from sklearn.model_selection import train_test_split
from sklearn.linear_model import Ridge
from typing import List, Tuple

def run_regression(df:pd.DataFrame, y_label:str, X_labels:List[str]) -> Tuple[np.array, np.array, np.array, np.array]:
    y_actual = df[y_label]
    X_data = df[X_labels]

    results = sm.OLS(y_actual, X_data).fit()

    y_pred = results.predict(X_data)
    coefficients = results.params.values
    p_values = results.pvalues.values

    return y_actual, y_pred, coefficients, p_values

def run_regression_with_test_split(df:pd.DataFrame, y_label:str, 
                                   X_labels:List[str], test_size:float=0.2) -> Tuple[np.array, np.array, np.array, np.array, np.array, np.array]:
    y_actual = df[y_label]
    X_data = df[X_labels]

    X_train, X_test, y_train, y_test = train_test_split(X_data, y_actual,
        test_size=test_size, random_state=0)

    results = sm.OLS(y_train, X_train).fit()

    y_test_pred = results.predict(X_test)
    y_pred = results.predict(X_data)
    coefficients = results.params.values
    p_values = results.pvalues.values
    
    return y_test, y_test_pred, coefficients, p_values, y_actual, y_pred

def run_ml_regression(df:pd.DataFrame, y_label:str, X_labels:List[str]) -> Tuple[np.array, np.array, np.array]:
    y_train = df[y_label]
    X_train = df[X_labels]

    results = Ridge(alpha=0.01, fit_intercept=False).fit(X_train, y_train)

    y_pred = results.predict(X_train)
    coefficients = results.coef_

    return y_train, y_pred, coefficients

def run_ml_regression_with_test_split(df:pd.DataFrame, y_label:str, X_labels:List[str], test_size:float=0.2) -> Tuple[np.array, np.array, np.array]:
    y_data = df[y_label]
    X_data = df[X_labels]

    X_train, X_test, y_train, y_test = train_test_split(X_data, y_data,
        test_size=test_size, random_state=0)

    results = Ridge(alpha=0.01, fit_intercept=False, random_state=0).fit(X_train, y_train)

    y_pred = results.predict(X_test)
    coefficients = results.coef_

    return y_test, y_pred, coefficients

def create_results_df(X_labels:List[str], coefficients:np.array, p_values:List[float]=None) -> pd.DataFrame:
    if p_values is None: # ml algo doesn't give p_values
        results_df = pd.DataFrame({'coefficient': coefficients}, index=X_labels)
    else:
        results_df = pd.DataFrame({'coefficient': coefficients, 'p_value': p_values}, index=X_labels)

    return results_df

def create_pred_df(df:pd.DataFrame, results_df:pd.DataFrame) -> pd.DataFrame:
    # calculate predictions
    X_labels = results_df.index
    pred_df = df[X_labels].copy()
    for x in X_labels:
        pred_df[x] = results_df['coefficient'].loc[x] * df[x]

    return pred_df


import numpy as np
import pandas as pd
import datetime as dt

def make_column_index(df:pd.DataFrame, column_label:str) -> None:
    df.index = df[column_label]
    df.drop(column_label, axis=1, inplace=True)
    df.index.name = None

def rename_column(df:pd.DataFrame, column_label:str, new_name:str) -> None:
    df.rename(columns={column_label: new_name}, inplace=True)

def remove_outliers(df:pd.DataFrame, column_label:str) -> str:
    raw_data = df[column_label]
    mean = np.mean(raw_data)
    std_dev = np.std(raw_data)
    outliers_cutoff = std_dev * 3
    lower_limit = mean - outliers_cutoff
    upper_limit = mean + outliers_cutoff

    no_outliers = raw_data.apply(lambda x: mean if x > upper_limit or x < lower_limit else x)

    outlier_column = f'{column_label} (-outliers)'
    df[outlier_column] = no_outliers
    return outlier_column

def unstack_data(df:pd.DataFrame, metric_column:str, unstack_column:str) -> pd.DataFrame:
    pivoted = pd.pivot_table(df, index=['date'], values=[metric_column], columns=[unstack_column], aggfunc=[np.sum])
    pivoted.columns = pivoted.columns.droplevel(0)
    pivoted.columns.name = None
    pivoted = pivoted.reset_index()
    pivoted.columns = [col[1] for col in pivoted.columns]

    metric_columns = list(pivoted.columns[1:])
    metric_columns = [f"{c} | {metric_column}" for c in metric_columns]

    pivoted.columns = ["date"] + metric_columns
    pivoted.fillna(0, inplace=True)

    return pivoted

def transpose_data(df:pd.DataFrame) -> pd.DataFrame:
    date_col = df.columns[0]
    df = df.T
    df.columns = df.iloc[0]
    df.drop(df.index[0], inplace=True)
    df.reset_index(inplace=True)
    df.rename(columns={"index": date_col}, inplace=True)
    df = df.rename_axis(None, axis = 1)
    return df

def interpolate_weekly_data(df, date_col=None, resample_col=None):
    df = df.copy()

    if date_col == None:
        date_col = df.columns[0]
        
    if resample_col == None:
        resample_col = df.columns[1]
        
    df[date_col] = df[date_col].apply(lambda x: dt.datetime.strptime(f"{x}-1", "%Y-%W-%w")) # mondays
    df[date_col] = pd.to_datetime(df[date_col]) # datetime
    df.set_index(date_col, inplace=True)
    df_reindexed = df.reindex(pd.date_range(start=df.index.min(),
                                        end=df.index.max() + dt.timedelta(days=6),
                                        freq='1D'))

    col_to_resample = df_reindexed.columns[0]
    df_reindexed[col_to_resample] = df_reindexed[col_to_resample].fillna(0)
    df_reindexed[col_to_resample] = df_reindexed[col_to_resample].astype(str)
    df_reindexed[col_to_resample] = df_reindexed[col_to_resample].apply(lambda x: x.replace(',',''))
    df_reindexed[col_to_resample] = df_reindexed[col_to_resample].apply(lambda x: x.replace('$',''))
    df_reindexed[col_to_resample] = df_reindexed[col_to_resample].apply(lambda x: x.replace('£',''))
    df_reindexed[col_to_resample] = df_reindexed[col_to_resample].apply(lambda x: x.replace('€',''))
    df_reindexed[col_to_resample] = pd.to_numeric(df_reindexed[col_to_resample])
    df_reindexed[col_to_resample].replace({0:np.nan}, inplace=True)
    df = df_reindexed.interpolate(method='linear')
    df = df / 7
    df.reset_index(inplace=True)
    df.rename({'index': 'date'}, axis=1, inplace=True)
    
    return df

def interpolate_monthly_data(df, date_col=None, resample_col=None):
    df = df.copy()

    if date_col == None:
        date_col = df.columns[0]
        
    if resample_col == None:
        resample_col = df.columns[1]

    df[date_col] = pd.to_datetime(df[date_col], format="%Y-%m")
    df['start_of_month'] = (df[date_col].dt.floor('d') + pd.offsets.MonthEnd(0) - pd.offsets.MonthBegin(1))
    df['end_of_month'] = pd.to_datetime(df['start_of_month']) + pd.offsets.MonthEnd(1)
    df['days_in_month'] = (df['end_of_month'] - df['start_of_month']).dt.days + 1
    df[resample_col] = df[resample_col] / df['days_in_month']
    reindexed = df.set_index("start_of_month")
    reindexed = reindexed.reindex(pd.date_range(start=reindexed.index.min(),
                                        end=reindexed.end_of_month.max(),
                                        freq='1D'))
    
    resampled = reindexed[resample_col]
    resampled.replace({0:np.nan}, inplace=True)
    resampled = resampled.interpolate(method='linear')
    resampled = resampled.reset_index()
    resampled.rename({'index': 'date'}, axis=1, inplace=True)
    resampled.fillna(0, inplace=True)
    return resampled

def group_weekly(df, date_col:str) -> pd.DataFrame:
    weekly = df.copy()
    weekly['week'] = weekly[date_col].dt.isocalendar().week
    weekly['year'] = weekly[date_col].dt.isocalendar().year
    weekly['year_week'] = weekly['year'].astype(str) + "-" + weekly['week'].astype(str)
    weekly = weekly.groupby('year_week').sum()
    weekly.drop(['week', 'year'], axis=1, inplace=True)
    weekly.reset_index(inplace=True)
    return weekly

def group_monthly(df, date_col:str) -> pd.DataFrame:
    monthly = df.copy()
    monthly['month'] = monthly[date_col].dt.month
    monthly['year'] = monthly[date_col].dt.isocalendar().year
    monthly['year_month'] = monthly['year'].astype(str) + "-" + monthly['month'].astype(str)
    monthly = monthly.groupby('year_month').sum()
    monthly.drop(['month', 'year'], axis=1, inplace=True)
    monthly.reset_index(inplace=True)
    return monthly

def handle_search_trends_data(df:pd.DataFrame) -> pd.DataFrame:
    # delete any '<' signs for low volume days
    for c in df.select_dtypes(include=['object']).columns[1:]:
        df[c] = df[c].str.replace('<', '')
        df[c] = pd.to_numeric(df[c])

    date_col = df.columns[0]
    df[date_col] = pd.to_datetime(df[date_col])
    df.set_index(date_col, inplace=True)
    df_reindexed = df.reindex(pd.date_range(start=df.index.min(),
                                            end=df.index.max() + dt.timedelta(days=6), freq='1D'))
    df = df_reindexed.interpolate(method='linear')
    df = df.round(1)
    df.reset_index(inplace=True)
    df.rename({'index': 'date'}, axis=1, inplace=True)
    return df

def handle_covid_data(data:pd.DataFrame, sub_region_1:str=None) -> pd.DataFrame:
    if sub_region_1 is None:
        df = data[data['sub_region_1'].isnull()]
    else:
        df = data[data['sub_region_1'] == sub_region_1]
        df = df[df['sub_region_2'].isnull()]
    
    df.reset_index(inplace=True)
    return df[df.columns[9:]]

def handle_weather_data(df:pd.DataFrame) -> pd.DataFrame:
    year = df['YEAR'].astype(str)
    month = df['MO'].astype(str)
    day = df['DY'].astype(str)
    
    month = month.apply(lambda x: '0'+x if len(x) == 1 else x)
    day = day.apply(lambda x: '0'+x if len(x) == 1 else x)
    
    df['date'] = pd.to_datetime(year + "-" + month + "-" + day)
    df = df[['date', 'T2M_RANGE', 'T2M_MAX', 'T2M_MIN', 'T2M']]
    
    return df

def create_holiday_dummies(df:pd.DataFrame) -> pd.DataFrame:
    dr = pd.date_range(start=df['date'].min(), end=df['date'].max())
    date_df = pd.DataFrame({'date': dr})
    for _, row in df.iterrows():
        date_df[row[1]] = (date_df['date'] == row[0])
        
    date_df.iloc[:, 1:] = date_df.iloc[:, 1:].astype(int)
    return date_df

def create_date_range_dummies(df:pd.DataFrame) -> pd.DataFrame:
    dr = pd.date_range(start=df['start'].min(), end=df['end'].max())
    
    date_df = pd.DataFrame({'date': dr})

    for _, row in df.iterrows():
        date_df[row[2]] = (date_df['date'] >= row[0]) & (date_df['date'] <= row[1])
        
    date_df.iloc[:, 1:] = date_df.iloc[:, 1:].astype(int)
    return date_df

def add_start_of_month(df:pd.DataFrame, date_col:str) -> pd.DataFrame:
    df['start_of_month'] = (df[date_col].dt.floor('d') + pd.offsets.MonthEnd(0) - pd.offsets.MonthBegin(1))
    return df

def add_end_of_month(df:pd.DataFrame, date_col:str) -> pd.DataFrame:
    df['end_of_month'] = pd.to_datetime(df[date_col]) + pd.offsets.MonthEnd(1)
    return df

def num_days_in_month(df:pd.DataFrame, date_col:str) -> pd.DataFrame:
    start_of_month = (df[date_col].dt.floor('d') + pd.offsets.MonthEnd(0) - pd.offsets.MonthBegin(1))
    end_of_month = pd.to_datetime(df[date_col]) + pd.offsets.MonthEnd(1)
    df['days_in_month'] = (end_of_month - start_of_month).dt.days + 1
    return df


import statsmodels.tsa as tsa
import pandas as pd
from typing import List, Tuple

from .select import get_cols_containing

def add_constant(df:pd.DataFrame) -> None:
    df['constant'] = 1

def add_adstocks(df:pd.DataFrame, column_label:str, adstock_rates:List[float]=None) -> List:
    if adstock_rates is None:
        adstock_rates = [round(i*0.1,1) for i in range(1,10)]

    added_columns = list()
    for ar in adstock_rates:
        ar_column = f"{column_label} AR={ar}"
        df[ar_column] = tsa.filters.filtertools.recursive_filter(df[column_label], ar)
        added_columns.append(ar_column)
    return added_columns

def add_diminishing_returns(df:pd.DataFrame, column_label:str, saturation_levels:List[float]=None) -> List:
    if saturation_levels is None:
        saturation_levels = [round(i*0.1,1) for i in range(1,10)]

    added_columns = list()
    for dr in saturation_levels:
        dr_column = f"{column_label} DR={dr}"
        df[dr_column] = df[column_label]**dr
        added_columns.append(dr_column)
    return added_columns

def add_lags(df:pd.DataFrame, column_label:str, lags:List[float]=None) -> List:
    if lags is None:
        lags = [1, 2, 3, 7, 14, 30, 60, 90, 180, 365]

    added_columns = list()
    for l in lags:
        l_column = f"{column_label} Lag={l}"
        df[l_column] = df[column_label].shift(l)
        df[l_column] = df[l_column].fillna(0)
        added_columns.append(l_column)
    return added_columns

def add_interaction_effect(df:pd.DataFrame, column_label_a:str, column_label_b:str) -> str:
    interaction_name = f'{column_label_a} x {column_label_b}'

    df[interaction_name] = df[column_label_a] * df[column_label_b]
    return interaction_name

def add_day_of_week_dummies(df:pd.DataFrame, date_label:str=None) -> Tuple[List[str], pd.DataFrame]:
    if date_label is None:
        dates_index = pd.to_datetime(df.index)
        date_label = '_date'
        df[date_label] = dates_index
        
    else:
        dates_index = pd.to_datetime(df[date_label])

    df['day_of_week'] = dates_index.dt.day_name()
    df['day_of_week'] = df['day_of_week'].str.lower()
    dummies = pd.get_dummies(df['day_of_week'])
        
    dummies[date_label] = dates_index
    
    df = pd.merge(df, dummies, left_on=date_label, right_on=date_label, how='left')
    
    df.drop(['day_of_week'], axis=1, inplace=True)
    
    df.drop(['_date'], axis=1, inplace=True) # in case we added it
    dummies.drop([date_label], axis=1, inplace=True)
    
    return list(dummies.columns), df

def add_month_of_year_dummies(df:pd.DataFrame, date_label:str=None) -> Tuple[List[str], pd.DataFrame]:
    if date_label is None:
        dates_index = pd.to_datetime(df.index)
        date_label = '_date'
        df[date_label] = dates_index
        
    else:
        dates_index = pd.to_datetime(df[date_label])

    df['month_of_year'] = dates_index.dt.month_name()
    df['month_of_year'] = df['month_of_year'].str.lower()
    
    dummies = pd.get_dummies(df['month_of_year'])
        
    dummies[date_label] = df[date_label]
    
    df = pd.merge(df, dummies, left_on=date_label, right_on=date_label, how='left')
    
    df.drop(['month_of_year'], axis=1, inplace=True)

    df.drop(['_date'], axis=1, inplace=True) # in case we added it
    dummies.drop([date_label], axis=1, inplace=True)
    
    return list(dummies.columns), df

def add_payday_dummies(df:pd.DataFrame, date_label:str) -> Tuple[str, pd.DataFrame]:
    payday_column = 'payday'
    df[payday_column] = df[date_label].apply(lambda x:1 if x.strftime('%d') in ('14','15','16','30','31','1','2') else 0)

    return payday_column, df

def categorize_campaigns(df:pd.DataFrame, containing:str) -> Tuple[str, pd.DataFrame]: 
    containing_cols = get_cols_containing(df, containing)

    agg_label = df[f'"{containing}" Agg']
    df[agg_label] = df[containing_cols].sum()

    return agg_label, df
